import owlready2 as owl
import os
import json
import re

from owlready2.namespace import Ontology
from owlready2 import Thing as Individual
from sparc_assemble.core.assembler.sparql_queries import find_all_tools
from sparc_assemble.core.assembler.assemble_utils import post_process_sparql_results

class KG:

    def __init__(self, ontology_file=None):
        if ontology_file:
            self._ontology_file = ontology_file
        else:
            self._ontology_file = os.path.join(os.path.dirname(__file__), "../resources/EDAM.owl")

        self._ontology = owl.get_ontology(self._ontology_file).load()

    def _populate_ontology(self, ontology: Ontology, individual_names: list[str], individuals: list[Individual],
                           class_data: list[dict]) -> list[Individual]:
        """
        Create a list of individuals for a specific key (either input or output) for the tool and add them in the knowledge
        graph if not existing.
        Args:
            ontology (Ontology): an owl ontology
            individual_names (list[str]): list of individual names existing in the knowledge graph
            individuals (list[Individual]):  list of individuals existing in the knowledge graph
            class_data (list[dict]): tool description stored in a dict format
        Returns:
            individuals_list (list[Individual]): list of individuals for the class specified in class_data
        """
        individuals_list = []
        for i in range(len(class_data)):
            # If individual already in KG add to list, else create and add to list
            if list(class_data[i].keys())[0] in individual_names:
                id_indiv = individual_names.index(list(class_data[i].keys())[0])
                tmp_individual = individuals[id_indiv]
            else:
                tmp_individual = ontology.data_0006(list(class_data[i].keys())[0])
                tmp_individual.label = list(class_data[i].keys())[0]
            individuals_list.append(tmp_individual)
        return individuals_list

    def _populate_ontology_by_model(self, ontology: Ontology, individual_names: list[str], individuals: list[Individual],
                           class_data: list[dict]) -> list[Individual]:
        """
        Create a list of individuals for a specific key (either input or output) for the tool and add them in the knowledge
        graph if not existing.
        Args:
            ontology (Ontology): an owl ontology
            individual_names (list[str]): list of individual names existing in the knowledge graph
            individuals (list[Individual]):  list of individuals existing in the knowledge graph
            class_data (list[dict]): model description stored in a dict format
        Returns:
            individuals_list (list[Individual]): list of individuals for the class specified in class_data
        """
        individuals_list = []
        for i in range(len(class_data)):
            id = class_data[i].get("id")
            name = class_data[i].get("name")
            label = id + '.' + name
            # If individual already in KG add to list, else create and add to list
            if label in individual_names:
                id_indiv = individual_names.index(label)
                tmp_individual = individuals[id_indiv]
            else:
                tmp_individual = ontology.data_0006(label)
                tmp_individual.label = label

            individuals_list.append(tmp_individual)

        return individuals_list

    def add_tools(self, tool_library=None):

        # Retrieve json files with tool description
        tools_descriptions = [os.path.join(tool_library, f) for f in os.listdir(tool_library) if
                              f.endswith('.json')]

        for description in tools_descriptions:
            # Retrieve individuals already in KG
            individuals = list(self._ontology.individuals())
            individual_names = [str(indiv_onto).removeprefix(self._ontology_file[:-3]) for indiv_onto in
                                individuals]

            # Read description
            with open(description, 'r') as read_description:
                data = json.load(read_description)
                keys = list(data.keys())
                for key in keys:
                    if key == 'Input':
                        # Create list of input individuals
                        input_individuals = self._populate_ontology(self._ontology, individual_names, individuals,
                                                                    data[key])

                    elif key == 'Output':
                        # Create list of output individuals
                        output_individuals = self._populate_ontology(self._ontology, individual_names, individuals,
                                                                     data[key])

                    elif key == 'Operation':
                        method_name = data[key]
                        method = self._ontology.operation_0004(method_name)

                # Link inputs and outputs to method
                method.has_input = input_individuals
                method.has_output = output_individuals

                # Add description, label and cwl file name
                # method.comment = data['comment'] Maybe for later for NLP
                method.label = method_name
                method.isDefinedBy = data['cwl']

    def list(self):
        query_results = find_all_tools(self._ontology)
        query_results = post_process_sparql_results(query_results)
        for index, tool in enumerate(query_results, start=1):
            print(f"\t {index}- {','.join(tool[1])} -> {tool[0]} -> {', '.join(tool[2])}")

    def add_biomodels_model(self, inputs, outputs, model_name, model_path):
        # Retrieve individuals already in KG
        individuals = list(self._ontology.individuals())
        individual_names = [str(indiv_onto).removeprefix(self._ontology_file[:-3]) for indiv_onto in
                            individuals]

        input_individuals = self._populate_ontology(self._ontology, individual_names, individuals, inputs)

        output_individuals = self._populate_ontology(self._ontology, individual_names, individuals, outputs)


        method = self._ontology.operation_0004(model_name)

        # Link inputs and outputs to method
        method.has_input = input_individuals
        method.has_output = output_individuals

        # Add description, label and file name
        # method.comment = data['comment'] Maybe for later for NLP
        method.label = model_name
        method.isDefinedBy = model_path

    def add_sparc_model(self, model_path=None):
        # Retrieve individuals already in KG
        individuals = list(self._ontology.individuals())
        individual_names = [str(indiv_onto).removeprefix(self._ontology_file[:-3]) for indiv_onto in
                            individuals]

        # Read description
        with open(model_path, 'r') as read_description:
            data = json.load(read_description)
            keys = list(data.keys())
            for key in keys:
                if key == 'input':
                    # Create list of input individuals
                    input_individuals = self._populate_ontology_by_model(self._ontology, individual_names, individuals,
                                                                data[key])

                elif key == 'output':
                    # Create list of output individuals
                    output_individuals = self._populate_ontology_by_model(self._ontology, individual_names, individuals,
                                                                 data[key].get("data"))

                elif key == 'simulation':
                    name = data[key]["opencor"]["resource"]
                    parts = re.split(r'[/\.]', name)
                    method_name = parts[-2]
                    method = self._ontology.operation_0004(method_name)

            # Link inputs and outputs to method
            method.has_input = input_individuals
            method.has_output = output_individuals

            # Add description, label and cwl file name
            # method.comment = data['comment'] Maybe for later for NLP
            method.label = method_name
            method.isDefinedBy = model_path

    def save(self, save_path):
        self._ontology.save(file=save_path)




import owlready2 as owl

from owlready2.namespace import Ontology


def find_methods_from_request(ontology: Ontology, request: str) -> list[list[str]]:
    """
    Query ontology to find methods with request including in comment. Return methods with their inputs and outputs
    grouped.
    For now: comment matching very strict. In the future: request should be close to method description (flexibility).
    Args:
        ontology (Ontology): an owl ontology
        request (str): request from the user
    Returns:
        List of ['method', 'input1,input2,...', 'output,...']
    """
    sparql_query = """PREFIX edam: <%s>
        SELECT ?label_operation (GROUP_CONCAT(?label_input; separator=", ") as ?inputs) (GROUP_CONCAT(?label_output; separator=", ") as ?outputs)
        {?operation rdf:type edam:operation_0004 .
        ?operation rdfs:label ?label_operation .
        ?operation edam:has_input ?input .
        ?input rdfs:label ?label_input .
        ?operation edam:has_output ?output .
        ?output rdfs:label ?label_output .
        FILTER REGEX(?label_output, ".*%s.*", "i")} 
        GROUP BY ?label_operation
        """ % (ontology.base_iri, request)
    return list(owl.default_world.sparql(sparql_query))


def find_inferred_inputs(ontology: Ontology, input_str: str) -> list[list[str]]:
    """
    Query ontology to find methods with output = input (i.e. find methods that can compute input). Return methods with
    their inputs and outputs grouped.
    Args:
        ontology: an owl ontology
        input_str (str): label of an input
    Returns:
        List of ['method', 'input1,input2,...', 'output,...']
    """
    sparql_query = """PREFIX edam: <%s>
                SELECT ?label_operation (GROUP_CONCAT(?label_input; SEPARATOR=", ") AS ?inputs) (GROUP_CONCAT(?label_output; separator=", ") as ?outputs)
                {?current_input rdfs:label '%s' .
                ?operation edam:has_output ?current_input .
                ?operation rdfs:label ?label_operation .
                ?operation edam:has_output ?output .
                ?operation edam:has_input ?input .
                ?input rdfs:label ?label_input .
                ?output rdfs:label ?label_output} 
                GROUP BY ?operation
                """ % (ontology.base_iri, input_str)
    return list(owl.default_world.sparql(sparql_query))


def find_method_cwl(ontology: Ontology, method: str) -> list[list[str]]:
    """
    Query ontology to retrieve cwl file for the corresponding method through 'isDefinedBy' property.
    Args:
        ontology: an owl ontology
        method (str): label of a method
    Returns:
        List of ['path to cwl file']

    """
    sparql_query = """PREFIX edam: <%s>
        SELECT ?cwl
        {edam:%s rdfs:isDefinedBy ?cwl .}
        """ % (ontology.base_iri, method)
    return list(owl.default_world.sparql(sparql_query))


def find_all_tools(ontology: Ontology) -> list[list[str]]:
    """
    Query ontology to find all methods. Return methods with their inputs and outputs grouped.
    Args:
        ontology: an owl ontology
    Returns:
        List of ['method', 'input1,input2,...', 'output,...']
    """
    sparql_query = """PREFIX edam: <%s>
        SELECT ?label_operation (GROUP_CONCAT(?label_input; separator=", ") as ?inputs) (GROUP_CONCAT(?label_output; separator=", ") as ?outputs)
        {?operation rdf:type edam:operation_0004 .
        ?operation rdfs:label ?label_operation .
        ?operation edam:has_input ?input .
        ?input rdfs:label ?label_input .
        ?operation edam:has_output ?output .
        ?output rdfs:label ?label_output} 
        GROUP BY ?label_operation
        """ % ontology.base_iri
    return list(owl.default_world.sparql(sparql_query))



from typing import Union

from sparc_assemble.core.assembler.sparql_queries import *
from sparc_assemble.core.assembler.workflow_utils import *

from owlready2.namespace import Ontology

class Assembler():
    """
    From a knowledge graph, prompt user for request, retrieve possible workflows, prompt user for workflow choice and
    create the corresponding workflow (cwl) and inputs (yaml) files.
    """
    def __init__(self, KG_path: str):
        self.KG_path = KG_path
        self.workflow_options_dict = None
        self.workflow = None

    def run(self, path_to_cwl_folder=None):
        """
        Main function of the program. Parse arguments, load ontology, display options to the user, prompt user for choice in
        a GUI or terminal. Run the chosen workflow.
        """
        # Retrieve ontology
        ontology = owl.get_ontology(self.KG_path).load()

        # Define request options
        request_options = [ind.name for ind in list(ontology.individuals()) if
                           ind.is_a[0] == ontology.data_0006 and ontology.is_output_of in ind.get_properties()]

        # Display request options and prompt for request
        selected_request = prompt_input_request_nlp(request_options)

        # Retrieve methods that answer the request and store all possible workflows in a global dict
        self._handle_request_choice(request_options[selected_request], ontology)

        # Prompt user for methods and workflows choice, save workflow and inputs values
        self._workflow_choice_and_save(workflow_options_dict, request_options[selected_request],
                                                              ontology, path_to_cwl_folder)



    def _handle_request_choice(self, selected_choice: str, ontology: Ontology) -> None:
        """
        Find methods that answer the request and store all possible workflows in a global dict.
        Args:
            selected_choice (str): request chosen by the user
            ontology (Ontology): an owl ontology
        Returns:
            None
        """
        methods = find_methods_from_request(ontology, selected_choice)
        global workflow_options_dict
        workflow_options_dict = self._create_options_workflow_dict(ontology, methods)

    def _create_options_workflow_dict(self, ontology: Ontology, methods: list[list[str]]) -> dict[
        str, dict[str, list[list[str]]]]:
        """
        Create a dict with keys (str)=method that answer the request and values (dict)= dict with keys (str)=input combination
        Args:
            ontology (Ontology): an owl ontology
            methods (list[list[str]]): list of ['method', 'input_1, input_2', 'output_1, output_2']
        Returns:
            res_dict (dict[dict[str, list[list[str]]]]): dict with keys (str)=method that answer the request and values (dict)= dict with keys
                                (str)=input combination and values (list(list[str]))=list representing the workflow used
        """
        res_dict = {}  # Initialize empty dict

        for method in methods:
            post_processed_queries = post_process_sparql_results([method])
            input_combination = [post_processed_queries[0][1]]
            input_txt = list_to_string(post_processed_queries[0][1])
            output_txt = list_to_string(post_processed_queries[0][2])
            method_txt = post_processed_queries[0][0]
            # Initialize dict with original input combination as key, and workflow as value
            workflows = {input_txt: [[input_txt, method_txt, output_txt]]}
            # Update dict with all possible input combinations and workflows
            self._retrieve_possible_workflows(ontology, post_processed_queries, input_combination, workflows)
            res_dict[post_processed_queries[0][0]] = workflows  # store workflows in dict with method as key

        return res_dict

    def _retrieve_possible_workflows(self, ontology: Ontology, queries_result: list[list[str, list[str], list[str]]],
                                    combination: list[str], workflows: dict) -> None:
        """
        Recursive function to retrieve all the possible input combinations and the workflow to get from the combination to
        the desired output.
        Args:
            ontology (Ontology): an owl ontology
            queries_result (list[list[str, list[str], list[str]]]): list(['method', ['input',...], ['output',...]]))
                                                                    post processed query result given an input
            combination (list(list[str]): list of list of strings corresponding to the possible input combination so far
            workflows (dict): dict with keys (str)=input combination and values (list(list[str]))=list representing the
                    workflow used with the corresponding input combination each sublist being ['input', 'method', 'output']
        Returns:
            None: modify combination and workflows on the fly
        """

        if queries_result:
            # Loop over the number of methods
            for i in range(len(queries_result)):
                # Loop over the number of inputs of each method
                for j in range(len(queries_result[i][1])):
                    current_input = queries_result[i][1][j]
                    # Find method(s) with output current_input
                    queries = find_inferred_inputs(ontology, current_input)
                    post_processed_queries = post_process_sparql_results(queries)
                    for query in post_processed_queries:
                        # Retrieve list where current_input existed and replace it with new input. Add it to combination.
                        list_containing_previous_input = find_sublist_containing_item(combination, current_input)
                        try:
                            for previous_inp in list_containing_previous_input:
                                index = previous_inp.index(current_input)
                                new_input = previous_inp[:index] + query[1] + previous_inp[index + 1:]
                                unique_new_input = check_uniqueness_sublist(new_input)
                                combination.append(unique_new_input)

                                # Add input combination to workflow dict with the corresponding methods to get to the
                                # desired output
                                input_txt = list_to_string(query[1])
                                output_txt = list_to_string(query[2])
                                workflows[list_to_string(unique_new_input)] = [
                                    [input_txt, query[0], output_txt]]  # query[0] = method
                                for previous_method in workflows[list_to_string(previous_inp)]:
                                    workflows[list_to_string(unique_new_input)].append(previous_method)
                        except Exception as e:
                            print(e)
                    self._retrieve_possible_workflows(ontology, post_processed_queries, combination, workflows)
            return ''
        # Stop condition: if no method found from the previous query
        else:
            return ''

    def _workflow_choice_and_save(self, options_dict: dict[str, dict[str, list[list[str]]]], request: str, ontology: Ontology,
                                  path_to_cwl_folder: str) -> Union[None | str, str]:
        """
        Prompt user to choose a method and an input combination plus inputs value of the corresponding combination.
        TODO: Only support choosing one method.
        Args:
            options_dict (dict[str, dict[str, list[list[str]]]]): key(str)=method that answer the request and values(dict)= dict
                        with keys(str)=input combination and values(list(list[str]))=list representing the workflow used
                        with the corresponding input combination each sublist being ['input', 'method', 'output']
            request (str): request typed by the user
            ontology (Ontology): an owl ontology
            interface (MainInterface): GUI interface
        Returns:
            chosen_option (str): method chosen
            inputs (list[str]): list of inputs values for the chosen inputs combination.
        """
        # Extract methods that answer the request
        options = list(options_dict.keys())

        display_options(request, options, options_dict)
        if path_to_cwl_folder:
            choice_method, choice_input = choose_options(options, options_dict)
            self._store_workflow_choice(choice_method, choice_input, options_dict, options)

            # Save workflow and inputs values respectively in Workflows (.cwl file) and Jobs (.yml file) folders
            wf_inputs, wf_name = save_workflow(ontology, self.workflow, path_to_cwl_folder)


    def _store_workflow_choice(self, choice_method: int, choice_input: int,
                              methods_dict: dict[str, dict[str, list[list[str]]]],
                              options: list[str]) -> None:
        """
        Retrieve the chosen workflow and store it in a global variable (workflow).
        Args:
            choice_method (int): index of the chosen method
            choice_input (int): index of the chosen input combination
            methods_dict (dict[str, dict[str, list[list[str]]]]): dict of methods that can compute request with their
            corresponding workflow options
            options (list[str]): list of methods that can compute request
        Returns:
            None: modify workflow on the fly
        """

        chosen_option = options[choice_method - 1]
        chosen_inputs = list(methods_dict[chosen_option].keys())[choice_input - 1]
        self.workflow = methods_dict[chosen_option][chosen_inputs]


"""
This class is a modified version of the WorkflowGenerator class from the scriptcwl library due to deprecated versions of yaml.
van der Zwaan, J. M., Andela, B., de Vries, B., Veen, L., van Haren, R., Verhoeven, S., & van der Ploeg, A. (2018).
scriptcwl 0.8.1 (Version 0.8.1) [Computer software]. https://doi.org/10.5281/zenodo.1160725
"""

from __future__ import print_function

import codecs
import copy
import os
import shutil
from functools import partial

import tempfile
import six
from ruamel.yaml.comments import CommentedMap
from ruamel.yaml import YAML
from io import StringIO

from scriptcwl.scriptcwl import load_cwl, quiet
from scriptcwl.step import python_name
from scriptcwl.library import StepsLibrary
from scriptcwl.reference import Reference

import warnings

# import cwltool.load_tool functions
with quiet():
    # all is quiet in this scope
    from cwltool.main import print_pack

warnings.simplefilter('always', DeprecationWarning)


class WorkflowGenerator(object):
    """Class for creating a CWL workflow.

    The WorkflowGenerator class allows users to tie together inputs and outputs
    of the steps that need to be executed to perform a data processing task.
    The steps (i.e., command line tools and subworkflows) must be added to the
    steps library of the WorkflowGenerator object before they can be added to
    the workflow. To add steps to the steps library, the `load` method can be
    called with either a path to a directory containing CWL files:
    ::

        from scriptcwl import WorkflowGenerator

        with WorkflowGenerator() as wf:
            wf.load(steps_dir='/path/to/dir/with/cwl/steps/')

    Or a single CWL file:
    ::

        with WorkflowGenerator() as wf:
            wf.load(step_file='/path/to/cwl/step/file')

    ``wf.load()`` can be called multiple times. Step files are added to the
    steps library one after the other. For every step that is added to the
    steps library, a method with the same name is added to the
    WorkflowGenerator object. To add a step to the workflow, this method must
    be called (examples below).

    Next, the user should add one or more workflow inputs:
    ::
      txt_dir = wf.add_input(txt_dir='Directory')

    The ``add_input()`` method expects a ``name=type`` pair as input parameter.
    The pair connects an input name (``txt_dir`` in the example) to a CWL type
    (``'Directory'``). Optionally, a default value can be specified using
    ``default=value``.

    The ``add_input()`` method returns a string containing the name
    that can be used to connect this input parameter to step input parameter
    names.

    Next, workflow steps can be added. To add a workflow step, its method must
    be called on the WorkflowGenerator object. This method expects a list of
    (key, value) pairs as input parameters. (To find out what inputs a step
    needs call ``wf.inputs(<step name>)``. This method prints all the inputs
    and their types.) The method returns a list of strings containing output
    names that can be used as input for later steps, or that can be connected
    to workflow outputs.

    For example, to add a step called ``frog-dir`` to the workflow, the
    following method must be called:
    ::

        frogout = wf.frog_dir(dir_in=txt_dir)

    In a next step, ``frogout`` can be used as input:
    ::
        saf = wf.frog_to_saf(in_files=frogout)
        txt = wf.saf_to_txt(in_files=saf)

    Etcetera.

    When all steps of the workflow have been added, the user can specify
    workflow outputs:
    ::

        wf.add_outputs(txt=txt)

    Finally, the workflow can be saved to file:
    ::

        wf.save('workflow.cwl')

    To list steps and signatures available in the steps library, call:
    ::

        wf.list_steps()
    """

    def __init__(self, steps_dir=None, working_dir=None):
        self.working_dir = working_dir
        if self.working_dir:
            self.working_dir = os.path.abspath(self.working_dir)
            if not os.path.exists(self.working_dir):
                os.makedirs(self.working_dir)
        self.wf_steps = CommentedMap()
        self.wf_inputs = CommentedMap()
        self.wf_outputs = CommentedMap()
        self.step_output_types = {}
        self.steps_library = StepsLibrary(working_dir=working_dir)
        self.has_workflow_step = False
        self.has_scatter_requirement = False
        self.has_multiple_inputs = False

        self._wf_closed = False

        self.load(steps_dir)

    def __enter__(self):
        self._wf_closed = False

        return self

    def __exit__(self, *args):
        self.wf_steps = None
        self.wf_inputs = None
        self.wf_outputs = None
        self.step_output_types = None
        self.steps_library = None
        self.has_workflow_step = None
        self.has_scatter_requirement = None
        self.working_dir = None

        self._wf_closed = True

    def __getattr__(self, name, **kwargs):
        name = self.steps_library.python_names2step_names.get(name, None)
        step = self._get_step(name)
        return partial(self._make_step, step, **kwargs)

    def __str__(self):
        # use absolute paths for printing
        return yaml2string(self,
                           pack=False,
                           relpath=None,
                           wd=False)

    def _closed(self):
        if self._wf_closed:
            raise ValueError('Operation on closed WorkflowGenerator.')

    def load(self, steps_dir=None, step_file=None, step_list=None):
        """Load CWL steps into the WorkflowGenerator's steps library.

        Adds steps (command line tools and workflows) to the
        ``WorkflowGenerator``'s steps library. These steps can be used to
        create workflows.

        Args:
            steps_dir (str): path to directory containing CWL files. All CWL in
                the directory are loaded.
            step_file (str): path to a file containing a CWL step that will be
                added to the steps library.
        """
        self._closed()

        self.steps_library.load(steps_dir=steps_dir, step_file=step_file,
                                step_list=step_list)

    def list_steps(self):
        """Return string with the signature of all steps in the steps library.
        """
        self._closed()

        return self.steps_library.list_steps()

    def _has_requirements(self):
        """Returns True if the workflow needs a requirements section.

        Returns:
            bool: True if the workflow needs a requirements section, False
                otherwise.
        """
        self._closed()

        return any([self.has_workflow_step, self.has_scatter_requirement,
                   self.has_multiple_inputs])

    def inputs(self, name):
        """List input names and types of a step in the steps library.

        Args:
            name (str): name of a step in the steps library.
        """
        self._closed()

        step = self._get_step(name, make_copy=False)
        return step.list_inputs()

    def _add_step(self, step):
        """Add a step to the workflow.

        Args:
            step (Step): a step from the steps library.
        """
        self._closed()

        self.has_workflow_step = self.has_workflow_step or step.is_workflow
        self.wf_steps[step.name_in_workflow] = step

    def add_input(self, **kwargs):
        """Add workflow input.

        Args:
            kwargs (dict): A dict with a `name: type` item
                and optionally a `default: value` item, where name is the
                name (id) of the workflow input (e.g., `dir_in`) and type is
                the type of the input (e.g., `'Directory'`).
                The type of input parameter can be learned from
                `step.inputs(step_name=input_name)`.

        Returns:
            inputname

        Raises:
            ValueError: No or multiple parameter(s) have been specified.
        """
        self._closed()

        def _get_item(args):
            """Get a single item from args."""
            if not args:
                raise ValueError("No parameter specified.")
            item = args.popitem()
            if args:
                raise ValueError("Too many parameters, not clear what to do "
                                 "with {}".format(kwargs))
            return item

        symbols = None
        input_dict = CommentedMap()

        if 'default' in kwargs:
            input_dict['default'] = kwargs.pop('default')
        if 'label' in kwargs:
            input_dict['label'] = kwargs.pop('label')
        if 'symbols' in kwargs:
            symbols = kwargs.pop('symbols')

        name, input_type = _get_item(kwargs)

        if input_type == 'enum':
            typ = CommentedMap()
            typ['type'] = 'enum'
            # make sure symbols is set
            if symbols is None:
                raise ValueError("Please specify the enum's symbols.")
            # make sure symbols is not empty
            if symbols == []:
                raise ValueError("The enum's symbols cannot be empty.")
            # make sure the symbols are a list
            if type(symbols) != list:
                raise ValueError('Symbols should be a list.')
            # make sure symbols is a list of strings
            symbols = [str(s) for s in symbols]

            typ['symbols'] = symbols
            input_dict['type'] = typ
        else:
            # Set the 'type' if we can't use simple notation (because there is
            # a default value or a label)
            if bool(input_dict):
                input_dict['type'] = input_type

        msg = '"{}" is already used as a workflow input. Please use a ' +\
              'different name.'
        if name in self.wf_inputs:
            raise ValueError(msg.format(name))

        # Add 'type' for complex input types, so the user doesn't have to do it
        if isinstance(input_type, dict):
            input_dict['type'] = input_type

        # Make sure we can use the notation without 'type' if the input allows
        # it.
        if bool(input_dict):
            self.wf_inputs[name] = input_dict
        else:
            self.wf_inputs[name] = input_type

        return Reference(input_name=name)

    def add_outputs(self, **kwargs):
        """Add workflow outputs.

        The output type is added automatically, based on the steps in the steps
        library.

        Args:
            kwargs (dict): A dict containing ``name=source name`` pairs.
                ``name`` is the name of the workflow output (e.g.,
                ``txt_files``) and source name is the name of the step that
                produced this output plus the output name (e.g.,
                ``saf-to-txt/out_files``).
        """
        self._closed()

        for name, source_name in kwargs.items():
            obj = {}
            obj['outputSource'] = source_name
            obj['type'] = self.step_output_types[source_name]
            self.wf_outputs[name] = obj

    def set_documentation(self, doc):
        """Set workflow documentation.

        Args:
            doc (str): documentation string.
        """
        self._closed()

        self.documentation = doc

    def set_label(self, label):
        """Set workflow label.

        Args:
            label (str): short description of workflow.
        """
        self._closed()

        self.label = label

    def _get_step(self, name, make_copy=True):
        """Return step from steps library.

        Optionally, the step returned is a deep copy from the step in the steps
        library, so additional information (e.g., about whether the step was
        scattered) can be stored in the copy.

        Args:
            name (str): name of the step in the steps library.
            make_copy (bool): whether a deep copy of the step should be
                returned or not (default: True).

        Returns:
            Step from steps library.

        Raises:
            ValueError: The requested step cannot be found in the steps
                library.
        """
        self._closed()

        s = self.steps_library.get_step(name)
        if s is None:
            msg = '"{}" not found in steps library. Please check your ' \
                  'spelling or load additional steps'
            raise ValueError(msg.format(name))
        if make_copy:
            s = copy.deepcopy(s)
        return s

    def _generate_step_name(self, step_name):
        name = step_name
        i = 1

        while name in self.steps_library.step_ids:
            name = '{}-{}'.format(step_name, i)
            i += 1

        return name

    def to_obj(self, wd=False, pack=False, relpath=None):
        """Return the created workflow as a dict.

        The dict can be written to a yaml file.

        Returns:
            A yaml-compatible dict representing the workflow.
        """
        self._closed()

        obj = CommentedMap()
        obj['cwlVersion'] = 'v1.0'
        obj['class'] = 'Workflow'
        try:
            obj['doc'] = self.documentation
        except (AttributeError, ValueError):
            pass
        try:
            obj['label'] = self.label
        except (AttributeError, ValueError):
            pass
        if self._has_requirements():
            obj['requirements'] = []
        if self.has_workflow_step:
            obj['requirements'].append(
                {'class': 'SubworkflowFeatureRequirement'})
        if self.has_scatter_requirement:
            obj['requirements'].append({'class': 'ScatterFeatureRequirement'})
        if self.has_multiple_inputs:
            obj['requirements'].append(
                {'class': 'MultipleInputFeatureRequirement'})
        obj['inputs'] = self.wf_inputs
        obj['outputs'] = self.wf_outputs

        steps_obj = CommentedMap()
        for key in self.wf_steps:
            steps_obj[key] = self.wf_steps[key].to_obj(relpath=relpath,
                                                       pack=pack,
                                                       wd=wd)
        obj['steps'] = steps_obj

        return obj

    def to_script(self, wf_name='wf'):
        """Generated and print the scriptcwl script for the currunt workflow.

        Args:
            wf_name (str): string used for the WorkflowGenerator object in the
                generated script (default: ``wf``).
        """
        self._closed()

        script = []

        # Workflow documentation
        # if self.documentation:
        #    if is_multiline(self.documentation):
        #        print('doc = """')
        #        print(self.documentation)
        #        print('"""')
        #        print('{}.set_documentation(doc)'.format(wf_name))
        #    else:
        #        print('{}.set_documentation(\'{}\')'.format(wf_name,
        #        self.documentation))

        # Workflow inputs
        params = []
        returns = []
        for name, typ in self.wf_inputs.items():
            params.append('{}=\'{}\''.format(name, typ))
            returns.append(name)
        script.append('{} = {}.add_inputs({})'.format(
            ', '.join(returns), wf_name, ', '.join(params)))

        # Workflow steps
        returns = []
        for name, step in self.wf_steps.items():
            pyname = step.python_name
            returns = ['{}_{}'.format(pyname, o) for o in step['out']]
            params = ['{}={}'.format(name, python_name(param))
                      for name, param in step['in'].items()]
            script.append('{} = {}.{}({})'.format(
                ', '.join(returns), wf_name, pyname, ', '.join(params)))

        # Workflow outputs
        params = []
        for name, details in self.wf_outputs.items():
            params.append('{}={}'.format(
                name, python_name(details['outputSource'])))
        script.append('{}.add_outputs({})'.format(wf_name, ', '.join(params)))

        return '\n'.join(script)

    @staticmethod
    def _get_input_type(step, input_name):
        input_type = step.input_types.get(input_name)
        if not input_type:
            input_type = step.optional_input_types[input_name]

        if step.is_scattered:
            for scattered_input in step.scattered_inputs:
                if scattered_input == input_name:
                    input_type += '[]'

        return input_type

    def _get_source_type(self, ref):
        if isinstance(ref, list):
            self.has_multiple_inputs = True
            return [self._get_source_type_single(r) for r in ref]
        else:
            return self._get_source_type_single(ref)

    def _get_source_type_single(self, ref):
        if ref.refers_to_step_output():
            step = self.wf_steps[ref.step_name]
            return step.output_types[ref.output_name]
        else:
            input_def = self.wf_inputs[ref.input_name]
            if isinstance(input_def, six.string_types):
                return input_def
            return input_def['type']

    @staticmethod
    def _types_match(type1, type2):
        """Returns False only if it can show that no value of type1
        can possibly match type2.

        Supports only a limited selection of types.
        """
        if isinstance(type1, six.string_types) and \
                isinstance(type2, six.string_types):
            type1 = type1.rstrip('?')
            type2 = type2.rstrip('?')
            if type1 != type2:
                return False

        return True

    def _type_check_reference(self, step, input_name, reference):
        input_type = self._get_input_type(step, input_name)
        source_type = self._get_source_type(reference)
        if isinstance(source_type, list):
            # all source_types must be equal
            if len(set(source_type)) > 1:
                inputs = ['{} ({})'.format(n, t)
                          for n, t in zip(reference, source_type)]
                msg = 'The types of the workflow inputs/step outputs for ' \
                      '"{}" are not equal: {}.'.format(input_name,
                                                       ', '.join(inputs))
                raise ValueError(msg)

            # continue type checking using the first item from the list
            source_type = source_type[0]
            input_type = input_type['items']
            reference = reference[0]

        if self._types_match(source_type, input_type):
            return True
        else:
            if step.is_scattered:
                scattered = ' (scattered)'
            else:
                scattered = ''
            if reference.refers_to_wf_input():
                msg = 'Workflow input "{}" of type "{}" is not'
                msg += ' compatible with{} step input "{}" of type "{}"'
                msg = msg.format(
                        reference.input_name, source_type,
                        scattered,
                        python_name(input_name), input_type)
            else:
                msg = 'Step output "{}" of type "{}" is not'
                msg += ' compatible with{} step input "{}" of type "{}"'
                msg = msg.format(
                        reference, source_type,
                        scattered,
                        python_name(input_name), input_type)
            raise ValueError(msg)

    def _make_step(self, step, **kwargs):
        self._closed()

        for k in step.get_input_names():
            p_name = python_name(k)
            if p_name in kwargs.keys():
                if isinstance(kwargs[p_name], Reference):
                    step.set_input(p_name, six.text_type(kwargs[p_name]))
                elif isinstance(kwargs[p_name], list):
                    if all(isinstance(n, Reference) for n in kwargs[p_name]):
                        step.set_input(p_name, kwargs[k])
                    else:
                        raise ValueError(
                            'List of inputs contains an input with an '
                            'incorrect type for keyword argument {} (should '
                            'be a value returned by set_input or from adding '
                            'a step).'.format(p_name))
                else:
                    raise ValueError(
                        'Incorrect type (should be a value returned'
                        'by set_inputs() or from adding a step) for keyword '
                        'argument {}'.format(p_name))
            elif k not in step.optional_input_names:
                raise ValueError(
                    'Expecting "{}" as a keyword argument.'.format(p_name))

        if 'scatter' in kwargs.keys() or 'scatter_method' in kwargs.keys():
            # Check whether 'scatter' keyword is present
            if not kwargs.get('scatter'):
                raise ValueError('Expecting "scatter" as a keyword argument.')

            # Check whether the scatter variables are valid for this step
            scatter_vars = kwargs.get('scatter')
            if isinstance(scatter_vars, six.string_types):
                scatter_vars = [scatter_vars]

            for var in scatter_vars:
                if var not in step.get_input_names():
                    msg = 'Invalid variable "{}" for scatter.'
                    raise ValueError(msg.format(var))
                step.scattered_inputs.append(var)

            # Check whether 'scatter_method' keyword is present if there is
            # more than 1 scatter variable
            if not kwargs.get('scatter_method') and len(scatter_vars) > 1:
                msg = 'Expecting "scatter_method" as a keyword argument.'
                raise ValueError(msg)

            # Check validity of scatterMethod
            scatter_methods = ['dotproduct', 'nested_crossproduct',
                               'flat_crossproduct']
            m = kwargs.get('scatter_method')
            if m and m not in scatter_methods:
                msg = 'Invalid scatterMethod "{}". Please use one of ({}).'
                raise ValueError(msg.format(m, ', '.join(scatter_methods)))
            step.scatter_method = m

            # Update step output types (outputs are now arrays)
            for name, typ in step.output_types.items():
                step.output_types[name] = {'type': 'array', 'items': typ}

            self.has_scatter_requirement = True
            step.is_scattered = True

        # Check types of references
        for k in step.get_input_names():
            p_name = python_name(k)
            if p_name in kwargs.keys():
                self._type_check_reference(step, k, kwargs[p_name])

        # Make sure the step has a unique name in the workflow (so command line
        # tools can be added to the same workflow multiple times).
        name_in_wf = self._generate_step_name(step.name)
        step._set_name_in_workflow(name_in_wf)
        self.steps_library.step_ids.append(name_in_wf)

        # Create a reference for each output for use in subsequent
        # steps' inputs.
        outputs = []
        for n in step.output_names:
            ref = step.output_reference(n)
            self.step_output_types[ref] = step.output_types[n]
            outputs.append(ref)

        self._add_step(step)

        if len(outputs) == 1:
            return outputs[0]
        return outputs

    def validate(self):
        """Validate workflow object.

        This method currently validates the workflow object with the use of
        cwltool. It writes the workflow to a tmp CWL file, reads it, validates
        it and removes the tmp file again. By default, the workflow is written
        to file using absolute paths to the steps.
        """
        # define tmpfile
        (fd, tmpfile) = tempfile.mkstemp()
        os.close(fd)
        try:
            # save workflow object to tmpfile,
            # do not recursively call validate function
            self.save(tmpfile, mode='abs', validate=False)
            # load workflow from tmpfile
            document_loader, processobj, metadata, uri = load_cwl(tmpfile)
        finally:
            # cleanup tmpfile
            os.remove(tmpfile)

    def _pack(self, fname, encoding):
        """Save workflow with ``--pack`` option

        This means that al tools and subworkflows are included in the workflow
        file that is created. A packed workflow cannot be loaded and used in
        scriptcwl.
        """
        (fd, tmpfile) = tempfile.mkstemp()
        os.close(fd)
        try:
            self.save(tmpfile, mode='abs', validate=False)
            document_loader, processobj, metadata, uri = load_cwl(tmpfile)
        finally:
            # cleanup tmpfile
            os.remove(tmpfile)

        with codecs.open(fname, 'wb', encoding=encoding) as f:
            f.write(print_pack(document_loader, processobj, uri, metadata))

    def _convert_commented_map(self, obj):
        """
        Recursively convert CommentedMap to a regular Python dictionary.
        """
        if isinstance(obj, CommentedMap):
            return {self._convert_commented_map(k): self._convert_commented_map(v) for k, v in obj.items()}
        elif isinstance(obj, dict):
            return {self._convert_commented_map(k): self._convert_commented_map(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [self._convert_commented_map(i) for i in obj]
        elif isinstance(obj, str) or isinstance(obj, (int, float, bool, type(None))):
            # Primitive types and strings should be returned as-is
            return obj
        elif hasattr(obj, '__str__'):
            # Return the string representation of objects like `Reference`
            return str(obj)
        else:
            # Default: return the object as-is
            return obj

    def _yaml2string(self, wf, pack, relpath, wd):
        yaml = YAML(typ='unsafe', pure=True)  # Create a YAML instance
        yaml.default_flow_style = False  # Set the default flow style
        wf_obj = wf.to_obj(pack=pack, relpath=relpath, wd=wd)  # Convert the workflow to an object

        # Convert CommentedMap to regular dict
        wf_obj_dict = self._convert_commented_map(wf_obj)

        # Convert the object to a YAML string
        yaml_str = StringIO()
        yaml.dump(wf_obj_dict, yaml_str)
        yaml_content = yaml_str.getvalue()

        # Prepend the shebang line
        header = '#!/usr/bin/env cwl-runner\n'
        return header + yaml_content
        # s = [u'#!/usr/bin/env cwl-runner',
        #      yaml.dump(wf.to_obj(pack=pack, relpath=relpath, wd=wd),
        #                Dumper=yaml.RoundTripDumper)]
        # return u'\n'.join(s)

    def _save_yaml(self, fname, wf, pack, relpath, wd, encoding='utf-8'):
        with codecs.open(fname, 'wb', encoding=encoding) as yaml_file:
            yaml_file.write(self._yaml2string(wf=wf,
                                        pack=pack,
                                        relpath=relpath,
                                        wd=wd))

    def save(self, fname, mode=None, validate=True, encoding='utf-8',
             wd=False, inline=False, relative=False, pack=False):
        """Save the workflow to file.

        Save the workflow to a CWL file that can be run with a CWL runner.

        Args:
            fname (str): file to save the workflow to.
            mode (str): one of  (rel, abs, wd, inline, pack)
            encoding (str): file encoding to use (default: ``utf-8``).
        """
        self._closed()

        if mode is None:
            mode = 'abs'
            if pack:
                mode = 'pack'
            elif wd:
                mode = 'wd'
            elif relative:
                mode = 'rel'

            msg = 'Using deprecated save method. Please save the workflow ' \
                  'with: wf.save(\'{}\', mode=\'{}\'). Redirecting to new ' \
                  'save method.'.format(fname, mode)
            warnings.warn(msg, DeprecationWarning)

        modes = ('rel', 'abs', 'wd', 'inline', 'pack')
        if mode not in modes:
            msg = 'Illegal mode "{}". Choose one of ({}).'\
                  .format(mode, ','.join(modes))
            raise ValueError(msg)

        if validate:
            self.validate()

        dirname = os.path.dirname(os.path.abspath(fname))
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        if mode == 'inline':
            msg = ('Inline saving is deprecated. Please save the workflow '
                   'using mode=\'pack\'. Setting mode to pack.')
            warnings.warn(msg, DeprecationWarning)
            mode = 'pack'

        if mode == 'rel':
            relpath = dirname
            self._save_yaml(fname=fname, wf=self, pack=False, relpath=relpath,
                      wd=False)

        if mode == 'abs':
            self._save_yaml(fname=fname, wf=self, pack=False, relpath=None,
                      wd=False)

        if mode == 'pack':
            self._pack(fname, encoding)

        if mode == 'wd':
            if self.get_working_dir() is None:
                raise ValueError('Working directory not set.')
            else:
                # save in working_dir
                bn = os.path.basename(fname)
                wd_file = os.path.join(self.working_dir, bn)
                self._save_yaml(fname=wd_file, wf=self, pack=False, relpath=None,
                          wd=True)
                # and copy workflow file to other location (as though all steps
                # are in the same directory as the workflow)
                try:
                    shutil.copy2(wd_file, fname)
                except shutil.Error:
                    pass

    def get_working_dir(self):
        return self.working_dir

    def add_inputs(self, **kwargs):
        """Deprecated function, use add_input(self, **kwargs) instead.
        Add workflow input.

        Args:
            kwargs (dict): A dict with a `name: type` item
                and optionally a `default: value` item, where name is the
                name (id) of the workflow input (e.g., `dir_in`) and type is
                the type of the input (e.g., `'Directory'`).
                The type of input parameter can be learned from
                `step.inputs(step_name=input_name)`.

        Returns:
            inputname

        Raises:
            ValueError: No or multiple parameter(s) have been specified.
        """
        msg = ('The add_inputs() function is deprecation in favour of the '
               'add_input() function, redirecting...')
        warnings.warn(msg, DeprecationWarning)
        return self.add_input(**kwargs)



from owlready2.namespace import Ontology
from sparc_assemble.core.assembler.workflow import WorkflowGenerator
from scriptcwl.reference import Reference
from sparc_assemble.core.assembler.assemble_utils import *


def find_duplicate_inputs_name(wf_inputs: dict[str, list[str]]) -> dict[str, list[str]]:
    """
    Find inputs with same name between steps.
    Args:
        wf_inputs (dict[str, list[str]]): dict of inputs (values) for each step (key). Inputs are in the form of 'input_name: input_type'.
    Returns:
        dict of inputs with same name between steps. Key is the link name (step_n-step_n+1) and value is a list of inputs with same name between these steps.
    """
    list_inputs = list(wf_inputs.values())
    list_steps = list(wf_inputs.keys())
    duplicate_name_dict = {}
    for i in range(len(list_inputs) - 1):
        # Convert the lists to sets for efficient intersection
        set1 = set(list_inputs[i])
        tmp_list = list_inputs[i + 1:]
        for sublist in tmp_list:
            set2 = set(sublist)

            # Find the common elements
            common = set1.intersection(set2)

            # Convert the set back to a list if needed
            duplicate_name_dict[f'{list_steps[i]}-{list_steps[i + 1]}'] = list(common)

    return duplicate_name_dict


def modify_duplicate_inputs(duplicate_inputs: dict[str, list[str]], wf_inputs: dict[str, list[str]], inputs_match: dict[str, dict[str, str]]) -> None:
    """
    Modify duplicate inputs name according to user choices.
    Args:
        duplicate_inputs (dict[str, list[str]]): dict of inputs with same name between steps. Key is the link name (step_n-step_n+1) and value is a list of inputs with same name between these steps.
        wf_inputs (dict[str, list[str]]): dict of inputs (values) for each step (key). Inputs are in the form of 'input_name: input_type'.
        inputs_match (dict[str, dict[str, str]]): dict of name matches for inputs. Key is the step name and value is a dict of original input names (key) and corresponding name after linking (value).
    Returns:
        wf_inputs and inputs_match modified according to user choices.
    """
    # Count total number of duplicate inputs
    total_duplicate = sum(len(sublist) for sublist in duplicate_inputs.values())
    print(
        f'{total_duplicate} inputs are duplicated (same name and same type). \n Please select the ones that represent '
        f'the same input [y]. Otherwise [n], they will be indexed according to their step.')
    # Loop over steps with duplicate inputs
    for key in list(duplicate_inputs.keys()):
        # Loop over duplicate inputs between the 2 current steps
        for duplicate_inp in duplicate_inputs[key]:
            print(f'Input: {duplicate_inp} is duplicated between step {key.split("-")[0]} and {key.split("-")[1]}.')
            choice = str(input("Does it represent the same input for these two steps? "))
            # If duplicate inputs name represent the same input, remove the duplicated one from workflow inputs
            if choice.casefold() in ['yes', 'y']:
                step_2 = key.split('-')[1]
                wf_inputs[step_2].remove(duplicate_inp)
            # Else, modify their name by indexing them with their corresponding step name and update workflow inputs names
            else:
                step_1 = key.split('-')[0]
                step_2 = key.split('-')[1]
                modify_duplicate_name(duplicate_inp, wf_inputs, inputs_match, step_1, step_2)


def link_steps(outputs: dict[str, list[str]], inputs: dict[str, list[str]]) -> dict[str, list[str]]:
    """
    Link outputs from previous steps to inputs of the next step.
    Args:
        outputs (dict[str, list[str]]): dict of outputs (values) for each step (key). Outputs are in the form of 'output_name: output_type'.
        inputs (dict[str, list[str]]): dict of inputs (values) for each step (key). Inputs are in the form of 'input_name: input_type'.
    Returns:
        Dict of linked inputs and outputs for each pair of step linked. Format: dict{step_n-step_n+1: [("output_name: 'output_type'", "input_name: 'input_type'")]} for all linked steps
    """
    steps_name = list(inputs.keys())
    # Initialize lists and linked dict
    link_dict = {}
    # Loop over steps
    for index, method in enumerate(steps_name):
        # If first step, only store inputs as there is no previous step/outputs
        if index == 0:
            # Print current step and its inputs
            print(f'Step {index + 1} ({steps_name[index]}) inputs: {inputs[method]} \n')
        else:
            # Print current step and its inputs
            print(f'Step {index + 1} ({steps_name[index]}) inputs: {inputs[method]}')
            inputs_dict = {key.strip(): value.strip(" '") for item in inputs[method] for key, value in
                           [item.split(':', 1)]}  # dict{input_name: input_type} for current step
            # Retrieve outputs available from all previous steps: dict{step_name: ['output_name: output_type']}
            outputs_available = {prev_method: outputs[prev_method] for prev_method in steps_name[:index]}
            outputs_dict = {key: value for d in [
                {key.strip(): value.strip(" '") for item in outputs_available_list for key, value in
                 [item.split(':', 1)]} for outputs_available_list in outputs_available.values()] for key, value in
                            d.items()} # dict{output_name: output_type} for all previous steps
            print(f'Outputs available at this step: {outputs_available} \n')
            choice = str(input("Does any input of current step need to be linked to a previous output (yes/no): "))
            while choice.casefold() in ['yes', 'y']:
                input_to_link = str(input('\nInput name: ')).casefold()
                output_to_link = str(input('Output_name: ')).casefold()
                # Check if input and output provided are in available inputs/outputs
                if input_to_link in inputs_dict.keys() and output_to_link in outputs_dict.keys():
                    # Check type correspondence
                    if inputs_dict[input_to_link] == outputs_dict[output_to_link]:
                        key_steps = steps_name[index - 1] + '-' + steps_name[index]
                        # If link key already in dict, add new value
                        if key_steps in link_dict.keys():
                            value_steps = ('"' + output_to_link + ': ' + "'" + outputs_dict[output_to_link] + "'" + '"',
                                           '"' + input_to_link + ': ' + "'" + inputs_dict[input_to_link] + "'" + '"') # ("output_name: 'output_type'", "input_name: 'input_type'")
                            if value_steps in link_dict[key_steps]:
                                print('The link already exists.')
                            else:
                                link_dict[key_steps].append(value_steps)
                        # Else, create new link key and value
                        else:
                            value_steps = [(output_to_link + ': ' + "'" + outputs_dict[output_to_link] + "'",
                                            input_to_link + ': ' + "'" + inputs_dict[input_to_link] + "'")]
                            link_dict[key_steps] = value_steps
                        print('Done.')
                        choice = str(input('Is there another link to create? '))
                    else:
                        print('Input and output provided are not the same type.')
                        choice = str(input('Try again? '))
                else:
                    print('Input or output provided not in available inputs/outputs.')
                    choice = str(input('Try again? '))
    return link_dict # dict{step_n-step_n+1: [("output_name: 'output_type'", "input_name: 'input_type'")]} for all linked steps


def add_workflow_inputs(wf: WorkflowGenerator, wf_inputs: dict[str, list[str]]) -> dict[str, str]:
    """
    Add workflow inputs to the workflow object and store the final version of the workflow inputs in a dict.
    Args:
        wf (WorkflowGenerator): workflow object
        wf_inputs (dict[str, list[str]]): dict of inputs (values) for each step (key). Inputs are in the form of 'input_name: input_type'.
    Returns:
        global_inputs (dict[str, str]): dict of inputs for the workflow. Inputs are in the form of 'input_name: input_type'.
    """
    steps_names = wf_inputs.keys()
    global_inputs = {}
    # Loop over steps
    for name in steps_names:
        # Loop over inputs for each step
        for i in range(len(wf_inputs[name])):
            inp = {wf_inputs[name][i].split(':')[0]: wf_inputs[name][i].split(':', 1)[1][2:-1]} # dict{name:type}
            if '{' in wf_inputs[name][i].split(':', 1)[1][2:-1]: #handle special type such as File or Directory
                inp[wf_inputs[name][i].split(':')[0]] = eval(wf_inputs[name][i].split(':', 1)[1][2:-1])
            # Add input to workflow object
            wf.add_input(**inp) #TODO: this function can output reference for each input
            # Add input: dict{name:type} to global inputs dict
            global_inputs.update(inp)
    return global_inputs


def add_steps(wf: WorkflowGenerator, inputs_match: dict[str, dict[str, str]]) -> dict[str, list[Reference]]:
    """
    Add steps to the workflow object and store the final version of the workflow outputs in a dict.
    Args:
        wf (WorkflowGenerator): workflow object
        inputs_match (dict[str, dict[str, str]]): dict of name matches for inputs. Key is the step name and value is a dict of original input names (key) and corresponding name after linking (value).

    Returns:
        outputs_references (dict[str, list[Reference]]): dict of outputs for the workflow. Outputs are Reference.
    """
    outputs_references = {}
    # Loop over step
    for step, step_inputs in inputs_match.items():
        step_inputs_reference = {}
        # Loop over inputs for each step
        for key, value in step_inputs.items():
            if '/' in value:  # handle input linked to previous output
                step_inputs_reference[key] = Reference(step_name=value.split('/')[0], output_name=value.split('/')[1])
            else:
                step_inputs_reference[key] = Reference(input_name=value)
        make_step = wf.__getattr__(step, **step_inputs_reference)
        # Add make_step to outputs_references dict (make_step references the output)
        if step in outputs_references.keys():
            outputs_references[step].append(make_step())
        else:
            outputs_references[step] = [make_step()]
    return outputs_references


def add_workflow_outputs(wf: WorkflowGenerator, outputs_references: dict[str, list[Reference]]) -> None:
    """
    Add workflow outputs to the workflow object.
    Args:
        wf (WorkflowGenerator): workflow object
        outputs_references (dict[str, list[Reference]]): dict of outputs for the workflow. Outputs are Reference.
    Returns:
        None. wf is modified in place.
    """
    workflow_outputs = {str(value).split('/')[1]: value for key, values in outputs_references.items() for value in values} #TODO: handle same output name
    wf.add_outputs(**workflow_outputs)


def save_workflow(ontology: Ontology, workflow: list[list[str]], path_to_cwl_folder) -> tuple[dict, str]:
    """
    Create and save a CWL workflow from a list of steps (=workflow).
    Args:
        ontology (Ontology): an owl ontology
        workflow (list[list[str]]): list of [method, input, output] for each step of the workflow
    Returns:
        final_inputs (dict): dict of inputs for the workflow. Inputs are in the form of 'input_name: input_type'.
        workflow_name (str): name of the workflow
    """
    workflow_steps = find_steps_cwl(ontology, workflow, path_to_cwl_folder)
    # Create workflow object
    with WorkflowGenerator() as wf:
        # Load cwl file and store inputs and outputs in dicts with key = step name
        workflow_inputs, steps_outputs = retrieve_steps_information(wf, workflow_steps)
        # Link inputs and outputs between steps and store the links in a dict
        step_links = link_steps(steps_outputs, workflow_inputs)
        # Create dict of name matches for inputs. Key is the step name and value is a dict of original input names (key)
        # and corresponding name after linking (value).
        inputs_match = {key: {value.split(':', 1)[0]: value.split(':', 1)[0] for value in values} for key, values in workflow_inputs.items()}
        # Remove inputs link to previous outputs from workflow inputs and update inputs_match
        simplify_wf_inputs(workflow_inputs, step_links, inputs_match)  # TODO: something will go wrong for input_match if partially simplified and then change name
        # Find inputs with same name between steps and modify duplicate inputs name according to user choices
        duplicate_inputs_name = find_duplicate_inputs_name(workflow_inputs)
        if duplicate_inputs_name:
            modify_duplicate_inputs(duplicate_inputs_name, workflow_inputs, inputs_match)
        # Add workflow inputs, steps and outputs
        final_inputs = add_workflow_inputs(wf, workflow_inputs)
        out_ref = add_steps(wf, inputs_match)
        add_workflow_outputs(wf, out_ref)
        # Save workflow to a cwl file with name provided by user
        workflow_name = str(input('Workflow name: '))  # Get workflow name
        wf.save(f'workflows/{workflow_name}.cwl', mode='abs')  # TODO: remove hard coded path
    return final_inputs, workflow_name



import io

from setuptools import setup, find_packages
from pathlib import Path


def read(*paths, **kwargs):
    """Read the contents of a text file safely.
    >>> read("package_name", "VERSION")
    >>> read("README.md")
    ...
    """

    current_dir = Path(__file__).resolve().parent
    file_path = current_dir.joinpath(*paths)
    with io.open(
            file_path,
            encoding=kwargs.get("encoding", "utf8"),
    ) as open_file:
        content = open_file.read().strip()

    return content


def read_requirements(path):
    return [
        line.strip()
        for line in read(path).split("\n")
        if not line.startswith(('"', "#", "-", "git+"))
    ]


setup(
    name="sparc_assemble",
    version="0.1.9",
    description='A Python tool to find and automatically assemble tools and workflows to process SPARC datasets in accordance with FAIR principles.',
    long_description=(Path(__file__).parent / "README.md").read_text(),
    long_description_content_type="text/markdown",
    author="Thiranja Prasad Babarenda Gamage, Chinchien Lin, Jiali Xu, Mathilde Verlyck, Max Dang Vu",
    email="psam012@aucklanduni.ac.nz, clin864@aucklanduni.ac.nz",
    license="Apache-2.0",
    packages=find_packages(),
    include_package_data=True,
    package_data={'': ['resources./*']},
    install_requires=read_requirements("requirements.txt"),
)



import os
import copy

from thefuzz import process
from ruamel import yaml as YAML
from sparc_assemble.core.assembler.workflow import WorkflowGenerator
from owlready2.namespace import Ontology
from sparc_assemble.core.assembler.sparql_queries import find_method_cwl
from pathlib import Path
from sentence_transformers import SentenceTransformer, util

model = SentenceTransformer('paraphrase-MiniLM-L6-v2')  # A lightweight model, more available options exist

def fuzzy_match(user_input, possible_requests, threshold=80):
    # Fuzzy matching to find the closest match
    match, score = process.extractOne(user_input, possible_requests)
    if score >= threshold:
        return match
    return None

def sbert_match(user_input, possible_requests):

    query_embedding = model.encode(user_input)
    request_embeddings = model.encode(possible_requests)

    similarities = [(req, util.cos_sim([query_embedding], [emb])[0][0]) for req, emb in
                    zip(possible_requests, request_embeddings)]
    # similarities = [(req, doc.similarity(nlp(req))) for req in possible_requests]
    best_match = max(similarities, key=lambda x: x[1])

    if best_match[1] > 0.6:  # Adjust the threshold as needed
        return best_match[0]
    return None

def prompt_input_request_nlp(options: list[str]) -> str:
    """
    Prompt the user for their parameter of interest and match it to existing options if possible.:
    Args:
        options (list[str]): list of individuals that can be requested
    Returns:
         output (int): request
    """
    while True:
        # Prompt the user for their request
        request = input(f"\nPlease specify which parameter you would like to derive: ")

        match = fuzzy_match(request, options)

        if match:
            choice = input(f"Did you mean '{match}'? (yes/no): ")
            if choice.casefold() in ['yes', 'y']:
                return options.index(match)
            else:
                match = sbert_match(request, options)
                if match:
                    choice = input(f"Did you mean '{match}'? (yes/no): ")
                    if choice.casefold() in ['yes', 'y']:
                        return options.index(match)
                else:
                    print("No exact match found. Here are the possible requests:")
                    for request in options:
                        print(f"- {request}")
        else:
            match = sbert_match(request, options)
            if match:
                choice = input(f"Did you mean '{match}'? (yes/no): ")
                if choice.casefold() in ['yes', 'y']:
                    return options.index(match)
            else:
                print("No exact match found. Here are the possible requests:")
                for request in options:
                    print(f"- {request}")


def prompt_input_request(options: list[str]) -> str:
    """
    Display the possible options to the user with the following format:
    1) Individual 1
    2) Individual 2
    ...
    And request the user to choose one of the options.
    Args:
        options (list[str]): list of individuals that can be requested
    Returns:
         output (int): request
    """
    # Display the options to the user
    print(f"\nThe following parameters can be computed:")

    # Display the available options: possible input combination + their corresponding workflow
    for index, option in enumerate(options, start=1):
        print(f"\t {index}. {option}")

    while True:
        output = input("Type your request number: ")
        if int(output) <= len(options):
            return int(output)-1
        else:
            print(f"Only {len(options)} options are available. Your request '{output}' can't be processed. Please select a "
                  f"valid option.")


def display_options(request: str, options: list[str], options_dict: dict[str, dict[str, list[list[str]]]]) -> None:
    """
    Display the options to the user with the following format:
    1) Method_req (=Method that can compute request):
        1. Input combination: input1, input2, ... (workflow 1)
            [Workflow steps]
            - input1 -> method -> output1, output2, ...
            - input2 -> method -> output1, output2, ...
        2. Input combination: input1, input2, ... (workflow 2)
            - input1 -> method -> output1, output2, ...
            - input2 -> method -> output1, output2, ...
    Args:
        request (str): request from the user
        options (list[str]): list of methods that can compute request
        options_dict (dict[str, dict[str, list[list[str]]]]): dict of methods that can compute request with their
        corresponding workflow options
    Returns:
        None: display the options in terminal
    """
    # Display the options to the user
    print(f"\nThe following method can compute {request}:")

    # Display the available options: possible input combination + their corresponding workflow
    for index, option in enumerate(options, start=1):
        print(f"{index}. {option}:")
        print(f'\t {len(options_dict[option].keys())} possible workflow(s):')
        for index_input, option_input in enumerate(options_dict[option].keys(), start=1):
            print(f"\t- {index_input}. Input: {option_input}")
            workflow = options_dict[option][option_input]
            for method in workflow:
                input_txt = ''
                for i, item in enumerate(method[0].split(', ')):
                    # Display necessary inputs in bold
                    if item in option_input.split(', '):
                        if i == 0:
                            input_txt += '\033[1m' + item + '\033[0m'
                        else:
                            input_txt += ', ' + '\033[1m' + item + '\033[0m'
                    else:
                        if i == 0:
                            input_txt += item
                        else:
                            input_txt += ', ' + item
                print(f"\t\t- {input_txt} -> {method[1]} -> {method[2]}")


def choose_options(options: list[str], dict_methods: dict[str, dict[str, list[list[str]]]]) -> tuple[int, int]:
    """
    Prompt user for method and workflow choice.
    Args:
        options: list of methods that can compute request
        dict_methods: dict of methods that can compute request with their
        corresponding workflow options
    Returns:
        choice_method (int): index of the chosen method
        choice_input (int): index of the chosen workflow
    """
    # Get user method choice
    while True:
        try:
            choice_method = int(input("Enter the method of your choice (number): "))
            if 1 <= choice_method <= len(options):
                break
            else:
                print("Invalid method choice. Please select a valid option.")
        except ValueError:
            print("Invalid input. Please enter a number.")

    # Get user workflow choice
    while True:
        try:
            choice_input = int(input("Enter the workflow of your choice (number): "))
            if 1 <= choice_input <= len(list(dict_methods[options[choice_method - 1]].keys())):
                break
            else:
                print("Invalid inputs choice. Please select a valid option.")
        except ValueError:
            print("Invalid input. Please enter a number.")
    return choice_method, choice_input


def post_process_sparql_results(sparql_results: list[list[str]]) -> list[list[str, list[str], list[str]]]:
    """
    Post-processing of sparql queries. Allow to store inputs and outputs concatenated by sparql into lists for future
    iteration over them.
    Args:
        sparql_results (list(list[str]): sparql queries with format
        [['method1', 'input1, input2', 'output1'], ['method2', 'input3', 'output2, output3']]
    Returns:
        post_processed_queries (list(list[str, list[str], list[str]])): post processed queries with format
        [['method1', ['input1','input2'], ['output1']], ['method2', ['input3'], ['output2' , 'output3']]]
    """

    post_processed_queries = copy.deepcopy(sparql_results)

    # Store inputs and outputs as list of strings
    for sublist in post_processed_queries:
        for i in range(1, 3):
            try:
                sublist[i] = sublist[i].split(', ')
            except AttributeError:
                sublist[i] = [sublist[i]]

        inputs = sublist[1]
        outputs = sublist[2]

        # Verify inputs and outputs are unique (could be remove if owlready2 support distinct in group_concat)
        sublist[1] = check_uniqueness_sublist(inputs)
        sublist[2] = check_uniqueness_sublist(outputs)

    return post_processed_queries


def list_to_string(list_of_strings: list[str]) -> str:
    """
    Transform a list of strings to one string with elements separated by a comma
    Args:
        list_of_strings (list[str]): list of strings
    Returns: one string with elements from the list separated by a comma
    """

    return ', '.join(list_of_strings)


def find_sublist_containing_item(list_of_lists: list[list[str]], item: str) -> list[list[str]]:
    """
    Check if an item is present in a list of lists and return the lists containing this item.
    TODO: only work for single item but output could be more than one (not in the current implementation but should be one day)
    Args:
        list_of_lists (list[list[str]]): list of list of strings
        item (str): item to check the presence of
    Returns:
        List of sublists that contain the input item
    """

    list_of_sublists = []
    for sublist in list_of_lists:
        if item in sublist:
            list_of_sublists.append(sublist)
    return list_of_sublists


def check_uniqueness_sublist(list_combination: list[str]) -> list[str]:
    """
    Check if every sublist of a list contains unique item. Return list of sublists with unique items.
    Args:
        list_combination (list[str]): list of strings
    Returns:
        List of unique strings in the same order as the input list
    """
    # If every item is unique, return the list
    if len(list_combination) == len(set(list_combination)):
        return list_combination
    # Else, remove duplicates
    else:
        unique_list = []
        [unique_list.append(item) for item in list_combination if item not in unique_list]
        return unique_list


def find_steps_cwl(ontology: Ontology, workflow: list[list[str]], path_to_cwl_folder: str) -> list[str]:
    """
    Query ontology to retrieve cwl file for the corresponding method through 'isDefinedBy' property.
    Args:
        ontology: an owl ontology
        workflow (list[list[str]]): list of [method, input, output]
    Returns:
        List of ['path to cwl file']
    """
    workflow_steps = []
    # Loop over the steps of the workflow
    for step in workflow:
        method_cwl = find_method_cwl(ontology, step[1])  # step[1] is the method name
        method_cwl_path = Path(os.path.join(path_to_cwl_folder, method_cwl[0][0]))
        workflow_steps.append(method_cwl_path)
    return workflow_steps


def retrieve_steps_information(wf: WorkflowGenerator, workflow_steps: list[str]) -> (dict[str, list[str]], dict[str, list[str]]):
    """
    Retrieve inputs and outputs for each step of the workflow.
    Args:
        wf (WorkflowGenerator): workflow object
        workflow_steps (list[str]): list of paths to cwl files
    Returns:
        steps_inputs (dict[str,list[str]]): dict of inputs (values) for each step (key). Inputs are in the form of 'input_name: input_type'.
        steps_outputs (dict[str,list[str]]): dict of outputs (values) for each step (key). Outputs are in the form of 'output_name: output_type'.
    """
    # Initialize dicts
    steps_inputs = {}
    steps_outputs = {}
    # Loop over the step (as path to cwl files)
    for step_path in workflow_steps:
        wf.load(step_file=step_path)  # load step into workflow object steps library
        # Retrieve step inputs and outputs from the library and store them in their corresponding dict.
        step_name = os.path.splitext(os.path.basename(step_path))[0]
        step_outputs_dict = wf.steps_library.get_step(step_name).output_types
        steps_outputs[step_name] = [f"{key}: '{value}'" for key, value in step_outputs_dict.items()]
        step_inputs_optional_dict = wf.steps_library.get_step(step_name).optional_input_types
        step_inputs_dict = wf.steps_library.get_step(step_name).input_types
        steps_inputs[step_name] = [f"{key}: '{value}'" for key, value in step_inputs_dict.items()] + [
            f"{key}: '{value}'" for key, value in step_inputs_optional_dict.items()]

    return steps_inputs, steps_outputs


def simplify_wf_inputs(wf_inputs: dict[str, list[str]], links: dict[str, list[(str, str)]], inputs_match: dict[str, dict[str, str]]) -> None:
    """
    Remove inputs that are linked to previous outputs from the workflow inputs.
    Args:
        wf_inputs (dict[str, list[str]]): dict of inputs (values) for each step (key). Inputs are in the form of 'input_name: input_type'.
        links (dict[str, list[(str, str)]]): dict of links between inputs and outputs. Key is the link name and value is a list of tuples (output, input). Inputs and outputs are in the form of 'input/output_name: input/output_type'.
        inputs_match (dict[str, dict[str, str]]): dict of name matches for inputs. Key is the step name and value is a dict of original input names (key) and corresponding name after linking (value).
    Returns:
        wf_inputs and inputs_match are modified in place.
    """
    for key in links.keys():
        step = key.split('-')[1]  # retrieve step name n+1 (the one with the input link to previous output)
        for value in links[key]:
            wf_inputs[step].remove(value[1])  # remove the input link to previous output from workflow inputs
            inputs_match[step][value[1].split(':', 1)[0]] = key.split('-')[0] + '/' + value[0].split(':', 1)[0]  # update inputs name in inputs_match


def modify_duplicate_name(duplicate_inp: str, wf_inputs: dict[str, list[str]], inputs_match: dict[str, dict[str, str]],
                          step_1: str, step_2: str) -> None:
    """
        Modify duplicate inputs name by indexing them with their corresponding step name and update workflow inputs names and inputs_match names.
    Args:
        duplicate_inp (str): duplicate input name with format 'input_name: input_type'
        wf_inputs (dict[str, list[str]]): dict of inputs (values) for each step (key). Inputs are in the form of 'input_name: input_type'.
        inputs_match (dict[str, dict[str, str]]): dict of name matches for inputs. Key is the step name and value is a dict of original input names (key) and corresponding name after linking (value).
        step_1 (str): step name n
        step_2 (str): step name n+1
    Returns:
         wf_inputs and inputs_match modified in place.
    """
    new_inp_name_step_1 = duplicate_inp.split(':', 1)[0] + f'_{step_1}:' + duplicate_inp.split(':', 1)[1]
    wf_inputs[step_1] = [new_inp_name_step_1 if item == duplicate_inp else item for item in
                         wf_inputs.get(step_1)]
    new_inp_name_step_2 = duplicate_inp.split(':', 1)[0] + f'_{step_2}:' + duplicate_inp.split(':', 1)[1]
    wf_inputs[step_2] = [new_inp_name_step_2 if item == duplicate_inp else item for item in
                         wf_inputs.get(step_2)]
    inputs_match[step_1][duplicate_inp.split(':', 1)[0]] = new_inp_name_step_1.split(':', 1)[0]
    inputs_match[step_2][duplicate_inp.split(':', 1)[0]] = new_inp_name_step_2.split(':', 1)[0]







import requests
from zipfile import ZipFile
import os
from urllib.parse import urlparse, parse_qs


class WorkflowHub:
    def __init__(self, zip_url):
        """
        Initialize the WorkflowHub with the provided configuration.

        :param config: Dictionary containing zip_url, local_zip_path, and output_directory.
        """
        self._zip_url = zip_url

    def download_zip_file(self, url: str, local_filename: str) -> None:
        """
        Downloads a ZIP file from a specified URL and saves it locally.

        :param url: URL of the ZIP file to download.
        :param local_filename: Local file path where the ZIP file will be saved.
        """

        try:
            # Download the file
            response = requests.get(url, stream=True)
            if response.status_code == 200:
                with open(local_filename, 'wb') as file:
                    for chunk in response.iter_content(chunk_size=8192):
                        file.write(chunk)
                print(f"ZIP file downloaded successfully and saved as '{local_filename}'.")
            else:
                print(f"Failed to download the file. Status code: {response.status_code}")
        except requests.RequestException as e:
            print(f"An error occurred while downloading the file. Error: {e}")


    def unzip_folder(self, zip_path: str, output_dir: str) -> None:
        """
        Extracts all files from a ZIP archive to a specified directory.

        :param zip_path: Path to the ZIP file.
        :param output_dir: Directory where the files will be extracted.
        """
        try:
            with ZipFile(zip_path, 'r') as zip_file:
                # Ensure output directory exists
                os.makedirs(output_dir, exist_ok=True)

                # Extract all files to the output directory
                zip_file.extractall(output_dir)

                print(f"All files from the ZIP archive have been extracted to '{output_dir}'.")
        except Exception as e:
            print(f"An error occurred while processing the ZIP archive. Error: {e}")

    def generate_zip_filename(self, url: str) -> str:
        """
        Generates a ZIP filename based on the URL.

        :param url: URL of the ZIP file.
        :return: Generated ZIP filename.
        """
        parsed_url = urlparse(url)
        query_params = parse_qs(parsed_url.query)

        # Extract workflow ID and version from URL parameters
        workflow_id = parsed_url.path.split('/')[2]  # Assuming the path structure is /workflows/{id}/ro_crate
        version = query_params.get('version', ['1'])[0]  # Default to '1' if not specified

        # Create a formatted ZIP filename
        zip_filename = f"workflow-{workflow_id}-{version}.crate"
        return zip_filename


if __name__ == "__main__":
    zip_url = "https://workflowhub.eu/workflows/776/ro_crate?version=1"
    workflow_hub = WorkflowHub(zip_url)

    # generate ZIP filename
    zip_filename = workflow_hub.generate_zip_filename(zip_url)
    local_zip_path = zip_filename+'.zip'

    # Download the ZIP file
    workflow_hub.download_zip_file(zip_url, local_zip_path)

    # Unzip the entire contents of the ZIP archive
    output_directory = zip_filename
    workflow_hub.unzip_folder(local_zip_path, output_directory)
